/*
 * Provides the voxelIterator class which allows iteration over 3D voxel
 * coordinates.
 */
#ifndef VOXEL_ITERATOR_H
#define VOXEL_ITERATOR_H

#include <iterator>
#include <cstdint>
#include <tuple>

namespace NEWIMAGE {

  /*
   * The voxelIterator class can be used to generate 3D voxel
   * coordinates for a NEWIMAGE::volume instance.  The volume
   * class has vbegin() and vend() functions which return
   * iterators to the first and last voxels in the volume.
   *
   * This allows code to iterate over voxels like so:
   *
   *     volume<float> vol;
   *     ...
   *     for (auto it = vol.vbegin(); it < vol.vend(); it++) {
   *         auto [x, y, z] = (*it);
   *         // do something with voxel xyz
   *     }
   *
   * The voxelIterator class can also be used with std::for_each
   * and std::transform, e.g.:
   *
   *     void do_something(std::tuple<int64_t, int64_t, int64_t> xyz) {
   *       ...
   *     }
   *
   *     float calculate_something(std::tuple<int64_t, int64_t, int64_t> xyz) {
   *       ...
   *       return val;
   *     }
   *
   *     std::for_each(vol.vbegin(), vol.vend(), do_something):
   *
   *     std::vector<float> result;
   *     std::transform(vol.vbegin(), vol.vend(), std::back_inserter(result),
   *                    calculate_something);
   *
   * IMPORTANT: Coordinates generated by the voxelIterator are only valid for
   * within-bounds voxels - if you de-reference a voxelIterator that refers
   * to a voxel before vol.vbegin() or after vol.vend() (inclusive), the
   * voxel coordinates will be invalid.
   */
  class voxelIterator {
  public:
    // Iterator traits allowing voxelIterator
    // to be used with STL function such as
    // std::for_each.
    //
    // The voxelIterator is basically a random
    // access iterator with the exception that it
    // doesn't implement the operator[] function.
    using iterator_category = std::bidirectional_iterator_tag;
    using difference_type   = int64_t;
    using value_type        = std::tuple<int64_t, int64_t, int64_t>;
    using pointer           = value_type*;
    using reference         = value_type&;

  private:

    // X/Y/Z voxel dimensions
    const size_t _xsz;
    const size_t _ysz;
    const size_t _zsz;

    // current 1D/flattened voxel index.
    int64_t _i;

    // Tuple containing current 3D coordinates.
    // Updated on demand, in the operator* and
    // operator-> functions.
    mutable std::tuple<int64_t, int64_t, int64_t> _current;

    // Called by operator* and operator->.
    // Updates the contents of the _current voxel.
    void _update_current() const {

      auto z =  _i / (_xsz * _ysz);
      auto y = (_i - z * _xsz * _ysz) / _xsz;
      auto x =  _i - z * _xsz * _ysz - y * _xsz;

      std::get<0>(_current) = x;
      std::get<1>(_current) = y;
      std::get<2>(_current) = z;
    }

  public:

    // Create a voxelIterator for a 3D volume of the given shape.
    voxelIterator(int64_t xsz, int64_t ysz, int64_t zsz, int64_t i=0)
      : _xsz(xsz), _ysz(ysz), _zsz(zsz), _i(i), _current({0, 0, 0})
    {}

    // Return a reference to a std::tuple containing the current
    // 3D voxel coordinates.
    const reference operator*() const {
      _update_current();
      return _current;
    }

    // Return a pointer to a std::tuple containing the current
    // 3D voxel coordinates.
    const pointer operator->() {
      _update_current();
      return &_current;
    }

    // Pre-fix increment/decrement
    voxelIterator& operator++() { _i++; return *this; }
    voxelIterator& operator--() { _i--; return *this; }

    // Increment/decrement by an arbitrary amount.
    voxelIterator& operator+=(int64_t i)  { _i += i; return *this; }
    voxelIterator& operator-=(int64_t i)  { _i += i; return *this; }

    // Post-fix increment/decrement
    voxelIterator  operator++(int) { auto tmp = (*this); ++(*this); return tmp; }
    voxelIterator  operator--(int) { auto tmp = (*this); --(*this); return tmp; }

    // Comparison operators
    friend bool operator==(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i == rhs._i; }

    friend bool operator!=(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i != rhs._i; }

    friend bool operator<(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i < rhs._i; }

    friend bool operator>(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i > rhs._i; }

    friend bool operator<=(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i <= rhs._i; }

    friend bool operator>=(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i >= rhs._i; }

    // Calculate difference between iterators
    friend int64_t operator-(const voxelIterator& lhs, const voxelIterator& rhs)
    { return lhs._i - rhs._i; }

    // Add/subtract offset to iterator
    friend voxelIterator operator-(const voxelIterator& lhs, int64_t rhs)
    { auto tmp = lhs; tmp._i -= rhs; return tmp; }

    friend voxelIterator operator+(const voxelIterator& lhs, int64_t rhs)
    { auto tmp = lhs; tmp._i += rhs; return tmp; }
  };
}

#endif
